@@ -27,10 +27,12 @@
 #include "ev-selection.h"
 #include "ev-page-cache.h"
 #include "ev-view-accessible.h"
+#include "ev-link-accessible.h"
 #include "ev-view-private.h"
 
-static void ev_view_accessible_text_iface_init   (AtkTextIface *iface);
-static void ev_view_accessible_action_iface_init (AtkActionIface *iface);
+static void ev_view_accessible_text_iface_init      (AtkTextIface      *iface);
+static void ev_view_accessible_action_iface_init    (AtkActionIface    *iface);
+static void ev_view_accessible_hypertext_iface_init (AtkHypertextIface *iface);
 
 enum {
 	ACTION_SCROLL_UP,
@@ -53,17 +55,24 @@ static const gchar *const ev_view_accessible_action_descriptions[] =
 };
 
 struct _EvViewAccessiblePrivate {
-	/* Action */
-	gchar *action_descriptions[LAST_ACTION];
-	guint action_idle_handler;
+	guint         current_page;
+
+	/* AtkAction */
+	gchar        *action_descriptions[LAST_ACTION];
+	guint         action_idle_handler;
 	GtkScrollType idle_scroll;
+
+	/* AtkText */
 	GtkTextBuffer *buffer;
-	guint current_page;
+
+	/* AtkHypertext */
+	GHashTable    *links;
 };
 
 G_DEFINE_TYPE_WITH_CODE (EvViewAccessible, ev_view_accessible, GTK_TYPE_CONTAINER_ACCESSIBLE,
-			 G_IMPLEMENT_INTERFACE (ATK_TYPE_TEXT, ev_view_accessible_text_iface_init);
-			 G_IMPLEMENT_INTERFACE (ATK_TYPE_ACTION, ev_view_accessible_action_iface_init);
+			 G_IMPLEMENT_INTERFACE (ATK_TYPE_TEXT, ev_view_accessible_text_iface_init)
+			 G_IMPLEMENT_INTERFACE (ATK_TYPE_ACTION, ev_view_accessible_action_iface_init)
+			 G_IMPLEMENT_INTERFACE (ATK_TYPE_HYPERTEXT, ev_view_accessible_hypertext_iface_init)
 	)
 
 static void
@@ -78,7 +87,10 @@ ev_view_accessible_finalize (GObject *object)
 		g_free (priv->action_descriptions [i]);
 	if (priv->buffer)
 		g_object_unref (priv->buffer);
+	if (priv->links)
+		g_hash_table_destroy (priv->links);
 
+	G_OBJECT_CLASS (ev_view_accessible_parent_class)->finalize (object);
 }
 
 static void
@@ -94,7 +106,8 @@ ev_view_accessible_initialize (AtkObject *obj,
 	atk_object_set_role (obj, ATK_ROLE_DOCUMENT_FRAME);
 }
 
-static void ev_view_accessible_class_init (EvViewAccessibleClass *klass)
+static void
+ev_view_accessible_class_init (EvViewAccessibleClass *klass)
 {
 	GObjectClass *object_class = G_OBJECT_CLASS (klass);
 	AtkObjectClass *atk_class = ATK_OBJECT_CLASS (klass);
@@ -105,7 +118,8 @@ static void ev_view_accessible_class_init (EvViewAccessibleClass *klass)
 	g_type_class_add_private (klass, sizeof (EvViewAccessiblePrivate));
 }
 
-static void ev_view_accessible_init (EvViewAccessible *accessible)
+static void
+ev_view_accessible_init (EvViewAccessible *accessible)
 {
 	accessible->priv = G_TYPE_INSTANCE_GET_PRIVATE (accessible, EV_TYPE_VIEW_ACCESSIBLE, EvViewAccessiblePrivate);
 }
@@ -139,10 +153,10 @@ ev_view_accessible_get_text_buffer (EvViewAccessible *accessible, EvView *view)
 	return priv->buffer;
 }
 
-static gchar*
+static gchar *
 ev_view_accessible_get_text (AtkText *text,
-			     gint    start_pos,
-			     gint    end_pos)
+			     gint     start_pos,
+			     gint     end_pos)
 {
 	GtkWidget *widget;
 	GtkTextIter start, end;
@@ -167,7 +181,7 @@ ev_view_accessible_get_text (AtkText *text,
 
 static gunichar
 ev_view_accessible_get_character_at_offset (AtkText *text,
-					    gint    offset)
+					    gint     offset)
 {
 	GtkWidget *widget;
 	GtkTextIter start, end;
@@ -197,100 +211,74 @@ ev_view_accessible_get_character_at_offset (AtkText *text,
 	return unichar;
 }
 
-static gchar*
-ev_view_accessible_get_text_before_offset (AtkText	    *text,
-					   gint		    offset,
-					   AtkTextBoundary  boundary_type,
-					   gint		    *start_offset,
-					   gint		    *end_offset)
+static gchar *
+ev_view_accessible_get_text_for_offset (EvViewAccessible *view_accessible,
+                                        gint              offset,
+                                        GailOffsetType    offset_type,
+                                        AtkTextBoundary   boundary_type,
+                                        gint             *start_offset,
+                                        gint             *end_offset)
 {
-	GtkWidget *widget;
-	gpointer layout = NULL;
-	GailTextUtil *gail_text = NULL;
-	gchar *retval = NULL;
-	GtkTextBuffer *buffer;
+        GtkWidget     *widget;
+        GtkTextBuffer *buffer;
+        GailTextUtil  *gail_text;
+        gchar         *retval;
 
-	widget = gtk_accessible_get_widget (GTK_ACCESSIBLE (text));
-	if (widget == NULL)
-		/* State is defunct */
-		return NULL;
+        widget = gtk_accessible_get_widget (GTK_ACCESSIBLE (view_accessible));
+        if (widget == NULL)
+                /* State is defunct */
+                return NULL;
 
-	buffer = ev_view_accessible_get_text_buffer (EV_VIEW_ACCESSIBLE (text), EV_VIEW (widget));
-	if (!buffer)
-		return NULL;
+        buffer = ev_view_accessible_get_text_buffer (view_accessible, EV_VIEW (widget));
+        if (!buffer)
+                return NULL;
 
-	gail_text = gail_text_util_new ();
-	gail_text_util_buffer_setup (gail_text, buffer);
-	retval = gail_text_util_get_text (gail_text, layout,
-	                                  GAIL_BEFORE_OFFSET, boundary_type,
-	                                  offset, start_offset, end_offset);
-	g_object_unref (gail_text);
+        gail_text = gail_text_util_new ();
+        gail_text_util_buffer_setup (gail_text, buffer);
+        retval = gail_text_util_get_text (gail_text, NULL, offset_type, boundary_type,
+                                          offset, start_offset, end_offset);
+        g_object_unref (gail_text);
 
-	return retval;
+        return retval;
 }
 
-static gchar*
-ev_view_accessible_get_text_at_offset (AtkText          *text,
-				       gint             offset,
-				       AtkTextBoundary  boundary_type,
-				       gint             *start_offset,
-				       gint             *end_offset)
+static gchar *
+ev_view_accessible_get_text_before_offset (AtkText        *text,
+                                           gint            offset,
+                                           AtkTextBoundary boundary_type,
+                                           gint           *start_offset,
+                                           gint           *end_offset)
 {
-	GtkWidget *widget;
-	gpointer layout = NULL;
-	GailTextUtil *gail_text = NULL;
-	gchar *retval = NULL;
-	GtkTextBuffer *buffer;
-
-	widget = gtk_accessible_get_widget (GTK_ACCESSIBLE (text));
-	if (widget == NULL)
-		/* State is defunct */
-		return NULL;
-
-	buffer = ev_view_accessible_get_text_buffer (EV_VIEW_ACCESSIBLE (text), EV_VIEW (widget));
-	if (!buffer)
-		return NULL;
-
-	gail_text = gail_text_util_new ();
-	gail_text_util_buffer_setup (gail_text, buffer);
-	retval = gail_text_util_get_text (gail_text, layout,
-	                                  GAIL_AT_OFFSET, boundary_type,
-	                                  offset, start_offset, end_offset);
-	g_object_unref (gail_text);
-
-	return retval;
+        return ev_view_accessible_get_text_for_offset (EV_VIEW_ACCESSIBLE (text),
+                                                       offset, GAIL_BEFORE_OFFSET,
+                                                       boundary_type,
+                                                       start_offset, end_offset);
 }
 
-static gchar*
-ev_view_accessible_get_text_after_offset (AtkText	    *text,
-					  gint		    offset,
-					  AtkTextBoundary   boundary_type,
-					  gint		    *start_offset,
-					  gint		    *end_offset)
+static gchar *
+ev_view_accessible_get_text_at_offset (AtkText        *text,
+                                       gint            offset,
+                                       AtkTextBoundary boundary_type,
+                                       gint           *start_offset,
+                                       gint           *end_offset)
 {
-	GtkWidget *widget;
-	gpointer layout = NULL;
-	GailTextUtil *gail_text = NULL;
-	gchar *retval = NULL;
-	GtkTextBuffer *buffer;
+        return ev_view_accessible_get_text_for_offset (EV_VIEW_ACCESSIBLE (text),
+                                                       offset, GAIL_AT_OFFSET,
+                                                       boundary_type,
+                                                       start_offset, end_offset);
+}
 
-	widget = gtk_accessible_get_widget (GTK_ACCESSIBLE (text));
-	if (widget == NULL)
-		/* State is defunct */
-		return NULL;
-
-	buffer = ev_view_accessible_get_text_buffer (EV_VIEW_ACCESSIBLE (text), EV_VIEW (widget));
-	if (!buffer)
-		return NULL;
-
-	gail_text = gail_text_util_new ();
-	gail_text_util_buffer_setup (gail_text, buffer);
-	retval = gail_text_util_get_text (gail_text, layout,
-	                                  GAIL_AFTER_OFFSET, boundary_type,
-	                                  offset, start_offset, end_offset);
-	g_object_unref (gail_text);
-
-	return retval;
+static gchar *
+ev_view_accessible_get_text_after_offset (AtkText        *text,
+                                          gint            offset,
+                                          AtkTextBoundary boundary_type,
+                                          gint            *start_offset,
+                                          gint           *end_offset)
+{
+        return ev_view_accessible_get_text_for_offset (EV_VIEW_ACCESSIBLE (text),
+                                                       offset, GAIL_AFTER_OFFSET,
+                                                       boundary_type,
+                                                       start_offset, end_offset);
 }
 
 static gint
@@ -408,58 +396,48 @@ ev_view_accessible_get_character_extents (AtkText      *text,
 					  AtkCoordType coords)
 {
 	GtkWidget *widget, *toplevel;
+	EvView *view;
 	EvRectangle *areas = NULL;
-	EvRectangle *rect = NULL;
+	EvRectangle *doc_rect;
 	guint n_areas = 0;
-	EvPageCache *page_cache;
-	gint x_widget, y_widget, x_window, y_window;
-	gdouble scale;
-	GtkBorder border;
-	GdkRectangle page_area;
+	gint x_widget, y_widget;
+	GdkRectangle view_rect;
 
 	widget = gtk_accessible_get_widget (GTK_ACCESSIBLE (text));
 	if (widget == NULL)
 		/* State is defunct */
 		return;
 
-	page_cache = EV_VIEW (widget)->page_cache;
-	if (!page_cache)
+	view = EV_VIEW (widget);
+	if (!view->page_cache)
 		return;
 
-	ev_view_get_page_extents (EV_VIEW (widget), EV_VIEW (widget)->current_page,
-	                          &page_area, &border);
-
-	scale = EV_VIEW (widget)->scale;
-	ev_page_cache_get_text_layout (page_cache, EV_VIEW (widget)->current_page, &areas, &n_areas);
-	if (!areas)
+	ev_page_cache_get_text_layout (view->page_cache, view->current_page, &areas, &n_areas);
+	if (!areas || offset >= n_areas)
 		return;
 
-	if (offset >= n_areas)
-		return;
-
-	rect = areas + offset;
-	*x = (int)(rect->x1 * scale);
-	*y = (int)(rect->y1 * scale);
-
-	*width = (int)(fabs (rect->x2 - rect->x1) * scale);
-	*height = (int)(fabs (rect->y2 - rect->y1) * scale);
+	doc_rect = areas + offset;
+	_ev_view_transform_doc_rect_to_view_rect (view, view->current_page, doc_rect, &view_rect);
+	view_rect.x -= view->scroll_x;
+	view_rect.y -= view->scroll_y;
 
 	toplevel = gtk_widget_get_toplevel (widget);
 	gtk_widget_translate_coordinates (widget, toplevel, 0, 0, &x_widget, &y_widget);
-	*x += x_widget;
-	*y += y_widget;
+	view_rect.x += x_widget;
+	view_rect.y += y_widget;
 
 	if (coords == ATK_XY_SCREEN) {
+		gint x_window, y_window;
+
 		gdk_window_get_origin (gtk_widget_get_window (toplevel), &x_window, &y_window);
-		*x += x_window;
-		*y += y_window;
+		view_rect.x += x_window;
+		view_rect.y += y_window;
 	}
 
-	*x -= EV_VIEW (widget)->scroll_x;
-	*y -= EV_VIEW (widget)->scroll_y;
-
-	*x += page_area.x;
-	*y += page_area.y;
+	*x = view_rect.x;
+	*y = view_rect.y;
+	*width = view_rect.width;
+	*height = view_rect.height;
 }
 
 static gint
@@ -469,14 +447,15 @@ ev_view_accessible_get_offset_at_point (AtkText      *text,
 					AtkCoordType coords)
 {
 	GtkWidget *widget, *toplevel;
+	EvView *view;
 	EvRectangle *areas = NULL;
 	EvRectangle *rect = NULL;
 	guint n_areas = 0;
-	guint i = 0;
-	EvPageCache *page_cache;
-	gint x_window, y_window, x_widget, y_widget;
-	gint offset=-1, rx, ry;
-	gdouble scale;
+	guint i;
+	gint x_widget, y_widget;
+	gint offset=-1;
+	GdkPoint view_point;
+	gdouble doc_x, doc_y;
 	GtkBorder border;
 	GdkRectangle page_area;
 
@@ -485,45 +464,36 @@ ev_view_accessible_get_offset_at_point (AtkText      *text,
 		/* State is defunct */
 		return -1;
 
-	page_cache = EV_VIEW (widget)->page_cache;
-	if (!page_cache)
+	view = EV_VIEW (widget);
+	if (!view->page_cache)
 		return -1;
 
-	ev_view_get_page_extents (EV_VIEW (widget), EV_VIEW (widget)->current_page,
-	                          &page_area, &border);
-
-	scale = EV_VIEW (widget)->scale;
-	ev_page_cache_get_text_layout (page_cache, EV_VIEW (widget)->current_page, &areas, &n_areas);
+	ev_page_cache_get_text_layout (view->page_cache, view->current_page, &areas, &n_areas);
 	if (!areas)
 		return -1;
 
-	rx = x;
-	ry = y;
-
-	rx -= page_area.x;
-	ry -= page_area.y;
-
-	rx += EV_VIEW (widget)->scroll_x;
-	ry += EV_VIEW (widget)->scroll_y;
-
+	view_point.x = x;
+	view_point.y = y;
 	toplevel = gtk_widget_get_toplevel (widget);
 	gtk_widget_translate_coordinates (widget, toplevel, 0, 0, &x_widget, &y_widget);
-	rx -= x_widget;
-	ry -= y_widget;
+	view_point.x -= x_widget;
+	view_point.y -= y_widget;
 
 	if (coords == ATK_XY_SCREEN) {
+		gint x_window, y_window;
+
 		gdk_window_get_origin (gtk_widget_get_window (toplevel), &x_window, &y_window);
-		rx -= x_window;
-		ry -= y_window;
+		view_point.x -= x_window;
+		view_point.y -= y_window;
 	}
 
-	rx /= scale;
-	ry /= scale;
+	ev_view_get_page_extents (view, view->current_page, &page_area, &border);
+	_ev_view_transform_view_point_to_doc_point (view, &view_point, &page_area, &doc_x, &doc_y);
 
 	for (i = 0; i < n_areas; i++) {
 		rect = areas + i;
-		if (rx >= rect->x1 && rx <= rect->x2 &&
-		    ry >= rect->y1 && ry <= rect->y2)
+		if (doc_x >= rect->x1 && doc_x <= rect->x2 &&
+		    doc_y >= rect->y1 && doc_y <= rect->y2)
 			offset = i;
 	}
 
@@ -557,9 +527,9 @@ ev_view_accessible_get_n_selections (AtkText *text)
 		return 0;
 }
 
-static gchar*
+static gchar *
 ev_view_accessible_get_selection (AtkText *text,
-				  gint    selection_num,
+				  gint     selection_num,
 				  gint    *start_pos,
 				  gint    *end_pos)
 {
@@ -592,8 +562,8 @@ ev_view_accessible_get_selection (AtkText *text,
 
 static gboolean
 ev_view_accessible_add_selection (AtkText *text,
-				  gint    start_pos,
-				  gint    end_pos)
+				  gint     start_pos,
+				  gint     end_pos)
 {
 	GtkWidget *widget;
 	GtkTextBuffer *buffer;
@@ -632,7 +602,7 @@ ev_view_accessible_add_selection (AtkText *text,
 
 static gboolean
 ev_view_accessible_remove_selection (AtkText *text,
-				     gint    selection_num)
+				     gint     selection_num)
 {
 	GtkWidget *widget;
 	GtkTextBuffer *buffer;
@@ -671,9 +641,9 @@ ev_view_accessible_remove_selection (AtkText *text,
 
 static gboolean
 ev_view_accessible_set_selection (AtkText *text,
-				  gint	  selection_num,
-				  gint    start_pos,
-				  gint    end_pos)
+				  gint	   selection_num,
+				  gint     start_pos,
+				  gint     end_pos)
 {
 	GtkWidget *widget;
 	GtkTextBuffer *buffer;
@@ -707,10 +677,9 @@ ev_view_accessible_set_selection (AtkText *text,
 	return retval;
 }
 
-static void ev_view_accessible_text_iface_init (AtkTextIface * iface)
+static void
+ev_view_accessible_text_iface_init (AtkTextIface * iface)
 {
-	g_return_if_fail (iface != NULL);
-
 	iface->get_text = ev_view_accessible_get_text;
 	iface->get_character_at_offset = ev_view_accessible_get_character_at_offset;
 	iface->get_text_before_offset = ev_view_accessible_get_text_before_offset;
@@ -728,7 +697,6 @@ static void ev_view_accessible_text_iface_init (AtkTextIface * iface)
 	iface->get_default_attributes = ev_view_accessible_get_default_attributes;
 	iface->get_character_extents = ev_view_accessible_get_character_extents;
 	iface->get_offset_at_point = ev_view_accessible_get_offset_at_point;
-	return;
 }
 
 static gboolean
@@ -745,7 +713,7 @@ ev_view_accessible_idle_do_action (gpointer data)
 
 static gboolean
 ev_view_accessible_action_do_action (AtkAction *action,
-				     gint      i)
+				     gint       i)
 {
 	EvViewAccessiblePrivate* priv = EV_VIEW_ACCESSIBLE (action)->priv;
 
@@ -778,7 +746,7 @@ ev_view_accessible_action_get_n_actions (AtkAction *action)
 
 static const gchar *
 ev_view_accessible_action_get_description (AtkAction *action,
-					   gint      i)
+					   gint       i)
 {
 	EvViewAccessiblePrivate* priv = EV_VIEW_ACCESSIBLE (action)->priv;
 
@@ -793,7 +761,7 @@ ev_view_accessible_action_get_description (AtkAction *action,
 
 static const gchar *
 ev_view_accessible_action_get_name (AtkAction *action,
-				    gint      i)
+				    gint       i)
 {
 	if (i < 0 || i >= LAST_ACTION)
 		return NULL;
@@ -803,7 +771,7 @@ ev_view_accessible_action_get_name (AtkAction *action,
 
 static gboolean
 ev_view_accessible_action_set_description (AtkAction   *action,
-					   gint        i,
+					   gint         i,
 					   const gchar *description)
 {
 	EvViewAccessiblePrivate* priv = EV_VIEW_ACCESSIBLE (action)->priv;
@@ -819,7 +787,8 @@ ev_view_accessible_action_set_description (AtkAction   *action,
 	return TRUE;
 }
 
-static void ev_view_accessible_action_iface_init (AtkActionIface * iface)
+static void
+ev_view_accessible_action_iface_init (AtkActionIface * iface)
 {
 	iface->do_action = ev_view_accessible_action_do_action;
 	iface->get_n_actions = ev_view_accessible_action_get_n_actions;
@@ -828,6 +797,117 @@ static void ev_view_accessible_action_iface_init (AtkActionIface * iface)
 	iface->set_description = ev_view_accessible_action_set_description;
 }
 
+static GHashTable *
+ev_view_accessible_get_links (EvViewAccessible *accessible,
+			      EvView           *view)
+{
+	EvViewAccessiblePrivate* priv = accessible->priv;
+
+	if (view->current_page == priv->current_page && priv->links)
+		return priv->links;
+
+	priv->current_page = view->current_page;
+
+	if (priv->links)
+		g_hash_table_destroy (priv->links);
+	priv->links = g_hash_table_new_full (g_direct_hash,
+					     g_direct_equal,
+					     NULL,
+					     (GDestroyNotify)g_object_unref);
+	return priv->links;
+}
+
+static AtkHyperlink *
+ev_view_accessible_get_link (AtkHypertext *hypertext,
+			     gint          link_index)
+{
+	GtkWidget        *widget;
+	EvView           *view;
+	GHashTable       *links;
+	EvMappingList    *link_mapping;
+	gint              n_links;
+	EvMapping        *mapping;
+	EvLinkAccessible *atk_link;
+
+	widget = gtk_accessible_get_widget (GTK_ACCESSIBLE (hypertext));
+	if (widget == NULL)
+		/* State is defunct */
+		return NULL;
+
+	view = EV_VIEW (widget);
+	if (!EV_IS_DOCUMENT_LINKS (view->document))
+		return NULL;
+
+	links = ev_view_accessible_get_links (EV_VIEW_ACCESSIBLE (hypertext), view);
+
+	atk_link = g_hash_table_lookup (links, GINT_TO_POINTER (link_index));
+	if (atk_link)
+		return atk_hyperlink_impl_get_hyperlink (ATK_HYPERLINK_IMPL (atk_link));
+
+	link_mapping = ev_page_cache_get_link_mapping (view->page_cache, view->current_page);
+	if (!link_mapping)
+		return NULL;
+
+	n_links = ev_mapping_list_length (link_mapping);
+	mapping = ev_mapping_list_nth (link_mapping, n_links - link_index - 1);
+	atk_link = ev_link_accessible_new (EV_VIEW_ACCESSIBLE (hypertext),
+					   EV_LINK (mapping->data),
+					   &mapping->area);
+	g_hash_table_insert (links, GINT_TO_POINTER (link_index), atk_link);
+
+	return atk_hyperlink_impl_get_hyperlink (ATK_HYPERLINK_IMPL (atk_link));
+}
+
+static gint
+ev_view_accessible_get_n_links (AtkHypertext *hypertext)
+{
+	GtkWidget     *widget;
+	EvView        *view;
+	EvMappingList *link_mapping;
+
+	widget = gtk_accessible_get_widget (GTK_ACCESSIBLE (hypertext));
+	if (widget == NULL)
+		/* State is defunct */
+		return 0;
+
+	view = EV_VIEW (widget);
+	if (!EV_IS_DOCUMENT_LINKS (view->document))
+		return 0;
+
+	link_mapping = ev_page_cache_get_link_mapping (view->page_cache, view->current_page);
+
+	return link_mapping ? ev_mapping_list_length (link_mapping) : 0;
+}
+
+static gint
+ev_view_accessible_get_link_index (AtkHypertext *hypertext,
+				   gint          offset)
+{
+	guint i;
+
+	for (i = 0; i < ev_view_accessible_get_n_links (hypertext); i++) {
+		AtkHyperlink *hyperlink;
+		gint          start_index, end_index;
+
+		hyperlink = ev_view_accessible_get_link (hypertext, i);
+		start_index = atk_hyperlink_get_start_index (hyperlink);
+		end_index = atk_hyperlink_get_end_index (hyperlink);
+
+		if (start_index <= offset && end_index >= offset)
+			return i;
+	}
+
+	return -1;
+}
+
+static void
+ev_view_accessible_hypertext_iface_init (AtkHypertextIface *iface)
+{
+	iface->get_link = ev_view_accessible_get_link;
+	iface->get_n_links = ev_view_accessible_get_n_links;
+	iface->get_link_index = ev_view_accessible_get_link_index;
+}
+
 AtkObject *
 ev_view_accessible_new (GtkWidget *widget)
 {
