@@ -207,6 +207,14 @@ struct _EvView {
 
 	/* Accessibility */
 	AtkObject *accessible;
+
+	/* Caret navigation */
+	gboolean caret_enabled;
+	gint     cursor_offset;
+	gint     cursor_page;
+	gboolean cursor_visible;
+	guint    cursor_blink_timeout_id;
+	guint    cursor_blink_time;
 };
 
 struct _EvViewClass {
@@ -252,6 +260,9 @@ void _ev_view_transform_doc_rect_to_view_rect (EvView       *view,
 					       int           page,
 					       EvRectangle  *doc_rect,
 					       GdkRectangle *view_rect);
+void _ev_view_get_selection_colors (EvView  *view,
+				    GdkRGBA *bg_color,
+				    GdkRGBA *fg_color);
 
 #endif  /* __EV_VIEW_PRIVATE_H__ */
 
